import 'dart:convert';
import 'dart:typed_data';

import 'package:http/http.dart';

import 'api_request_method.dart';
import 'api_error.dart';
import 'api_response.dart';

import 'api_exception.dart';
import 'dto/transfer_object.dart';
import 'utils.dart';

/// Interface for api endpoints
abstract class ApiEndpoint<TData> {
  Future<ApiResponse<TData>> request(
      String url, Map<String, dynamic>? urlParams, ApiRequestMethod method,
      {Map<String, String>? headers, Map<String, String>? query, dynamic body});
}

/// Represents an endpoint
/// TData should be chosen according to the data structure of the endpoint:
/// Plain -> String
/// Json -> Map<String, dynamic> / List<dynamic>
/// Byte Data -> UInt8List
/// class extending TransferObject
class _ApiEndpoint<TData> implements ApiEndpoint<TData> {
  final String urlPostfix;
  final bool requiresAuthentication;
  final DTOFactory? factory;

  const _ApiEndpoint(this.urlPostfix,
      {this.requiresAuthentication = false, this.factory});

  Future<ApiResponse<TData>> request(
      String url, Map<String, dynamic>? urlParams, ApiRequestMethod method,
      {Map<String, String>? headers,
      Map<String, String>? query,
      dynamic body}) async {
    StringBuffer urlBuffer = StringBuffer(url);
    urlBuffer.write(_replacePlaceholders(urlPostfix, urlParams ?? {}));

    if (query?.isNotEmpty ?? false) {
      urlBuffer.write(_buildQueryString(query!));
    }

    headers ??= {};
    headers['Content-type'] = 'application/json; charset=UTF-8';

    final response = await _dispatch(
        method, Uri.parse(urlBuffer.toString()), headers, JsonEncoder().convert(body));

    return handleResponse(response);
  }

  ApiResponse<TData> handleResponse(Response response) {
    return ApiResponse(response, factory);
  }

  Future<Response> _dispatch(ApiRequestMethod type, Uri url,
      Map<String, String>? headers, dynamic body) async {
    switch (type) {
      case ApiRequestMethod.GET:
        return await get(url, headers: headers);
      case ApiRequestMethod.POST:
        return await post(url, headers: headers, body: body);
      case ApiRequestMethod.PUT:
        return await put(url, headers: headers, body: body);
      case ApiRequestMethod.PATCH:
        return await patch(url, headers: headers, body: body);
      case ApiRequestMethod.DELETE:
        return await delete(url, headers: headers);
      default:
        throw ApiError('Invalid request type: $type');
    }
  }

  String _buildQueryString(Map<String, String> query) {
    if (query.isEmpty) {
      return '';
    }

    return '?' +
        query.entries
            .map((e) =>
                '${Uri.encodeQueryComponent(e.key)}=${Uri.encodeQueryComponent(e.value)}')
            .join('&');
  }

  static final RegExp _plExp = RegExp('(?<!\\\\)\\{([\\w-]*)}');

  static String _replacePlaceholders(
      String source, Map<String, dynamic> values) {
    //TODO maybe add encode method for values, not just toString
    return source.replaceAllMapped(_plExp, (match) {
      final key = match.group(1);
      if (key == null) {
        throw ApiException('Invalid placeholder-key: ${match.group(0)}');
      }

      if (values[key] == null) {
        throw ApiException('Missing value in url params: $key');
      }

      return Uri.encodeComponent(values[key].toString());
    });
  }
}

/// Plain-text endpoint (ApiEndpoint<String>)
class PlainApiEndpoint extends _ApiEndpoint<String> {
  const PlainApiEndpoint(String urlPostfix,
      {bool requiresAuthentication = false})
      : super(urlPostfix, requiresAuthentication: requiresAuthentication);
}

/// ByteData endpoint (ApiEndpoint<Uint8List>)
class ByteApiEndpoint extends _ApiEndpoint<Uint8List> {
  const ByteApiEndpoint(String urlPostfix,
      {bool requiresAuthentication = false})
      : super(urlPostfix, requiresAuthentication: requiresAuthentication);
}

/// Json-Map endpoint (ApiEndpoint<Map<String, dynamic>>)
class JsonMapApiEndpoint extends _ApiEndpoint<Map<String, dynamic>> {
  const JsonMapApiEndpoint(String urlPostfix,
      {bool requiresAuthentication = false})
      : super(urlPostfix, requiresAuthentication: requiresAuthentication);
}

/// Json-List endpoint (ApiEndpoint<List<dynamic>>)
class JsonListApiEndpoint extends _ApiEndpoint<List<dynamic>> {
  const JsonListApiEndpoint(String urlPostfix,
      {bool requiresAuthentication = false})
      : super(urlPostfix, requiresAuthentication: requiresAuthentication);
}

/// Json-Map endpoint (ApiEndpoint<Map<String, dynamic>>)
class MapApiEndpoint<T extends TransferObject> extends _ApiEndpoint<T> {
  const MapApiEndpoint(String urlPostfix, DTOFactory<T> factory,
      {bool requiresAuthentication = false})
      : super(urlPostfix,
            factory: factory, requiresAuthentication: requiresAuthentication);
}

/// Json-List endpoint (ApiEndpoint<List<dynamic>>)
class ListApiEndpoint<T extends TransferObject> extends _ApiEndpoint<List<T>> {
  const ListApiEndpoint(String urlPostfix, DTOFactory<T> factory,
      {bool requiresAuthentication = false})
      : super(urlPostfix,
            factory: factory, requiresAuthentication: requiresAuthentication);

  @override
  ApiResponse<List<T>> handleResponse(Response response) {
    return ListApiResponse(response, factory as DTOFactory<T>);
  }
}
