import 'dart:convert';
import 'dart:typed_data';

import '../../appbase.dart';
import 'api_error.dart';
import 'api_exception.dart';
import 'status_code_exception.dart';
import 'package:http/http.dart';

import 'utils.dart';

class ApiResponse<TResult> {
  final Uri? requestUrl;
  final int statusCode;
  final TResult data;

  ApiResponse(Response response, [DTOFactory? factory, TResult? data])
      : requestUrl = response.request?.url,
        statusCode = response.statusCode,
        data = data ?? _handleData<TResult>(response.bodyBytes, factory);

  void throwOnError() {
    if (statusCode >= 400) {
      throw StatusCodeException.fromResponse(this);
    }
  }

  static dynamic _handleData<T>(Uint8List bodyBytes, DTOFactory? factory) {
    if (bodyBytes.isEmpty) {
      return null;
    }

    //TODO encoding from response content-type
    final encoding = utf8;

    if (factory != null) {
      final obj = JsonDecoder().convert(encoding.decode(bodyBytes));
      if (obj is Map<String, dynamic>) {
        return factory(obj);
      } else {
        throw ApiException('Invalid response.');
      }
    } else if (T == String) {
      return encoding.decode(bodyBytes);
    } else if (T.toString() == 'Map<String, dynamic>' || T == List) {
      return JsonDecoder().convert(encoding.decode(bodyBytes)) as T;
    } else if (T == Uint8List) {
      return bodyBytes;
    }

    throw ApiError.invalidType(T);
  }
}

class ListApiResponse<TResult extends TransferObject> extends ApiResponse<List<TResult>> {
  ListApiResponse(Response response, DTOFactory<TResult> factory)
      : super(response, null, _handleListData(response.bodyBytes, factory));

  static List<T>? _handleListData<T extends TransferObject>(Uint8List bodyBytes, DTOFactory<T> factory) {
    if (bodyBytes.isEmpty) {
      return null;
    }

    //TODO encoding from response content-type
    final encoding = utf8;

    final obj = JsonDecoder().convert(encoding.decode(bodyBytes));
    if (obj is List) {
      return obj.map((e) => factory(e)).toList();
    } else if (obj is Map<String, dynamic>) {
      return [factory(obj)];
    }

    throw ApiException('Invalid response.');
  }
}
